package org.bdware.sc.event;

import org.bdware.sc.NodeCenterConn;
import org.bdware.sc.NodeCenterConn.NodeKey;
import org.bdware.sc.util.HashUtil;
import org.bdware.sc.util.JsonUtil;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import static org.bdware.sc.ContractManager.instance;

/**
 * @author Kaidong Wu
 */
public class EventCenter {
//    private static final Logger LOGGER = LogManager.getLogger(EventCenter.class);

    /**
     * get the nearest node to the topic in the hash function range
     *
     * @param topic the topic
     * @return id of the node
     */
    public String getCenterByTopic(String topic) {
        String[] centers = getCentersByTopic(topic, 1);
        if (null == centers) {
            return null;
        }
        return centers[0];
    }

    /**
     * get k nearest nodes to the topic in the hash function range
     *
     * @param topic the topic
     * @param k     the number of required node ids
     * @return ids of k nearest nodes
     */
    public String[] getCentersByTopic(String topic, int k) {
        NodeCenterConn nodeCenterConn = instance.nodeCenterConn;
        if (null == nodeCenterConn) {
            return null;
        }
        NodeKey[] nodes = nodeCenterConn.listNodes();
        if (nodes.length == 0) {
            return null;
        }
        if (nodes.length == 1) {
            return new String[]{nodeCenterConn.getNodeId(nodes[0].id)};
        }

        // get hash with enough length
        String hash = HashUtil.sha3ToFixedLen(topic, nodes[0].id.length());
        BigInteger biH = new BigInteger(hash, 16);

        // binary search, to find the nearest node with hash
        int l = 0, r = nodes.length - 1, m = 0,
                comL = biH.compareTo(nodes[l].biId), comR = nodes[r].biId.compareTo(biH),
                comM;
        String selected;
        do {
            if (comL < 1) {
                selected = nodes[l].id;
                break;
            }
            if (comR < 1) {
                selected = nodes[r].id;
                break;
            }
            if (l + 1 == r) {
                if (biH.subtract(nodes[l].biId).compareTo(nodes[r].biId.subtract(biH)) < 1) {
                    selected = nodes[l].id;
                } else {
                    selected = nodes[r].id;
                }
                break;
            }
            m = (l + r) >> 1;
            comM = biH.compareTo(nodes[m].biId);
            if (comM < 1) {
                r = m;
                comR = -comM;
            } else {
                l = m;
                comL = comM;
            }
        } while (true);
        List<String> ret = new ArrayList<>();
        ret.add(nodeCenterConn.getNodeId(selected));
        if (k > 1) {
            l = m - 1;
            r = m + 1;
            while (ret.size() < k && (l >= 0 || r < nodes.length)) {
                if (l < 0) {
                    ret.add(nodeCenterConn.getNodeId(nodes[r++].id));
                } else if (r >= nodes.length) {
                    ret.add(nodeCenterConn.getNodeId(nodes[l--].id));
                } else {
                    if (biH.subtract(nodes[l].biId).compareTo(nodes[r].biId.subtract(biH)) < 1) {
                        ret.add(nodeCenterConn.getNodeId(nodes[l--].id));
                    } else {
                        ret.add(nodeCenterConn.getNodeId(nodes[r++].id));
                    }
                }
            }
        }

        return ret.toArray(new String[0]);
    }

    /**
     * subscribe a topic in center
     *
     * @param topic     event topic
     * @param semantics event semantics, used to mark PRESUB events
     */
    public void subInCenter(String topic, REvent.REventSemantics semantics) {
        if (null == instance.nodeCenterConn) {
            return;
        }
        REvent msg = new REvent(
                topic,
                REvent.REventType.SUBSCRIBE,
                String.format("{\"subscriber\":\"%s\"}",
                        instance.nodeCenterConn.getNodeId(null)),
                "");
        msg.setSemantics(semantics);
        msg.doSignature(instance.nodeCenterConn.getNodeKeyPair());
        String nodeId = getCenterByTopic(topic);
        instance.nodeCenterConn.deliverEvent(JsonUtil.toJson(msg), nodeId);
    }

    /**
     * deliver an event to center
     *
     * @param topic event topic
     * @param event the event
     * @return if this node is the center, return false; otherwise, return true
     */
    public boolean deliverEvent(String topic, REvent event) {
        if (null == instance.nodeCenterConn) {
            return false;
        }
        String nodeId = getCenterByTopic(topic);
        return instance.nodeCenterConn.deliverEvent(JsonUtil.toJson(event), nodeId);
    }

    /**
     * publish an event to another node; used by NodeConsumer
     *
     * @param eStr   event string
     * @param target id of the target node
     */
    public void publishEvent(String eStr, String target) {
        if (null != instance.nodeCenterConn) {
            instance.nodeCenterConn.deliverEvent(eStr, target);
        }
    }
}